/*
    This file is part of the Cheap Matrix library
    Copyright (C) 2000  Nicolas Brodu
    nicolas.brodu@free.fr

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public License
    along with this library; see the file COPYING.LIB.  If not, write to
    the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
    Boston, MA 02111-1307, USA.
*/
#include "CM_evd.h"
#include "lapack_types.h"
#include "proto.h"

extern "C" {
// Eigensmurf decomposition
int PROTO(dgeev,DGEEV)(char *jobvl, char *jobvr, integer *n, doublereal *a,
	integer *lda, doublereal *wr, doublereal *wi, doublereal *vl,
	integer *ldvl, doublereal *vr, integer *ldvr, doublereal *work,
	integer *lwork, integer *info);
// symmetric eigen vector decomposition
int PROTO(dsyev,DSYEV)(char *jobz, char *uplo, integer *n, doublereal *a,
	integer *lda, doublereal *w, doublereal *work, integer *lwork,
	integer *info);
} // extern "C"

// Don't pollute the standard namespace with our definitions
namespace CheapMatrix {

// A 0 pointer argument means no calculation
static bool ptr_evd(const Matrix &A, const Vector *reValues,
	const Vector *imValues, const Matrix *L, const Matrix *R)
{
	if (reValues) {CHECK_NOT_REF(EVD,A,(*reValues))}
	if (imValues) {CHECK_NOT_REF(EVD,A,(*imValues))}
	if (L) {CHECK_NOT_REF(EVD,A,(*L))}
	if (R) {CHECK_NOT_REF(EVD,A,(*R))}
	if (reValues && imValues) {CHECK_NOT_REF(EVD,(*reValues),(*imValues))}
	if (reValues && L) {CHECK_NOT_REF(EVD,(*reValues),(*L))}
	if (reValues && R) {CHECK_NOT_REF(EVD,(*reValues),(*R))}
	if (imValues && L) {CHECK_NOT_REF(EVD,(*imValues),(*L))}
	if (imValues && R) {CHECK_NOT_REF(EVD,(*imValues),(*R))}
	if (L && R) {CHECK_NOT_REF(EVD,(*L),(*R))}
	
	char jobvl=(L)?'V':'N';
	char jobvr=(R)?'V':'N';
	integer N = A.cols();
	if (N != A.rows()) return false;
	CHECK_MEM_LEAD(EVD,A,lda)
	Vector* wr = 0, *wi = 0;
	const Vector* p_re = reValues;
	const Vector* p_im = imValues;
	if (!p_re) {
		wr = new Vector(N);
		assert(wr);
		p_re = wr;
	} else {
		CHECK_DIM(EVD,(*reValues),N,1)
		CHECK_MEM(EVD,(*reValues))
	}
	if (!p_im) {
		wi = new Vector(N);
		assert(wi);
		p_im = wi;
	} else {
		CHECK_DIM(EVD,(*imValues),N,1)
		CHECK_MEM(EVD,(*imValues))
	}
	integer leadL = N, leadR = N;
	if (R) {
		CHECK_DIM(EVD,(*R),N,N)
		CHECK_MEM_LEAD(EVD,(*R),ldr)
		leadR = ldr;
	}
	if (L) {
		CHECK_DIM(EVD,(*L),N,N)
		CHECK_MEM_LEAD(EVD,(*L),ldl)
		leadL = ldl;
	}
	ALLOCATE_WORK(EVD, 4*N)
	integer info = 0;

	PROTO(dgeev,DGEEV)(&jobvl, &jobvr, &N, &A(1,1), &lda, &(*p_re)(1),
		&(*p_im)(1), L ? &(*L)(1,1) : 0, &leadL, R ? &(*R)(1,1) : 0, &leadR,
		&work[0], &lwork, &info);
	
	A.release();
	if (wr) delete wr;
	else reValues->release();
	if (wi) delete wi;
	else imValues->release();
	if (L) L->release();
	if (R) R->release();
	delete work;
	return !info;
}

// A 0 pointer argument means no calculation
static bool sym_ptr_evd(const Matrix &A, const Vector *eigenValues,
	const Matrix *eigenVectors)
{
	if (eigenValues) {CHECK_NOT_REF(symmetric_EVD,A,(*eigenValues))}
	if (eigenVectors) {CHECK_NOT_REF(symmetric_EVD,A,(*eigenVectors))}
	if (eigenValues && eigenVectors) {
		CHECK_NOT_REF(symmetric_EVD,(*eigenValues),(*eigenVectors))
	}
	
	char jobz=(eigenVectors)?'V':'N';
	char uplo='U';
	integer N = A.cols();
	if (N != A.rows()) return false;
	CHECK_MEM_LEAD(symmetric_EVD,A,lda)
	Vector* w = 0;
	const Vector* p_w = eigenValues;
	if (!p_w) {
		w = new Vector(N);
		assert(w);
		p_w = w;
	} else {
		CHECK_DIM(symmetric_EVD,(*eigenValues),N,1)
		CHECK_MEM(symmetric_EVD,(*eigenValues))
	}
	if (eigenVectors) {
		CHECK_DIM(symmetric_EVD,(*eigenVectors),N,N)
		// don't load fully in mem since only A is used (and destroyed)
	}
	ALLOCATE_WORK(symmetric_EVD,3*N-1)
	integer info = 0;

	PROTO(dsyev,DSYEV)(&jobz, &uplo, &N, &A(1,1), &lda, &(*p_w)(1),
		&work[0], &lwork, &info);
	
	if (w) delete w;
	else eigenValues->release();
	if (eigenVectors) eigenVectors->copy(A);
	A.release();
	delete work;
	return !info;
}

bool EVD(const Matrix &A, const Vector& reValues,
	const Vector& imValues, const Matrix& L, const Matrix& R)
{
	return ptr_evd(A, &reValues, &imValues, &L, &R);
}
bool EVD(const Matrix &A, const Vector& reValues,
	const Vector& imValues, const Matrix& R)
{
	return ptr_evd(A, &reValues, &imValues, 0, &R);
}
bool EVD(const Matrix &A, const Vector& reValues,
	const Matrix& L, const Matrix& R)
{
	return ptr_evd(A, &reValues, 0, &L, &R);
}
bool EVD(const Matrix &A, const Vector& reValues,
	const Matrix& R)
{
	return ptr_evd(A, &reValues, 0, 0, &R);
}
bool EVD(const Matrix &A, const Matrix& L, const Matrix& R)
{
	return ptr_evd(A, 0, 0, &L, &R);
}
bool EVD(const Matrix &A, const Matrix& R)
{
	return ptr_evd(A, 0, 0, 0, &R);
}
bool EVD(const Matrix &A, const Vector& reValues,
	const Vector& imValues)
{
	return ptr_evd(A, &reValues, &imValues, 0, 0);
}
bool EVD(const Matrix &A, const Vector& reValues)
{
	return ptr_evd(A, &reValues, 0, 0, 0);
}
// cannot overload with the same signature
bool left_EVD(const Matrix &A, const Vector& reValues,
	const Vector& imValues, const Matrix& L)
{
	return ptr_evd(A, &reValues, &imValues, &L, 0);
}
bool left_EVD(const Matrix &A, const Vector& reValues,
	const Matrix& L)
{
	return ptr_evd(A, &reValues, 0, &L, 0);
}
bool left_EVD(const Matrix &A, const Matrix& L)
{
	return ptr_evd(A, 0, 0, &L, 0);
}

// Same thing, optimised assuming the matrix A is symmetric
bool symmetric_EVD(const Matrix &A, const Vector& eigenValues,
	const Matrix& eigenVectors)
{
	return sym_ptr_evd(A,&eigenValues,&eigenVectors);
}
bool symmetric_EVD(const Matrix &A, const Vector& eigenValues)
{
	return sym_ptr_evd(A,&eigenValues,0);
}
bool symmetric_EVD(const Matrix &A, const Matrix& eigenVectors)
{
	return sym_ptr_evd(A,0,&eigenVectors);
}
}
/*
//Test
main()
{
	Matrix A(3,3), L(3,3), R(3,3);
	Vector re(3), im(3);

	A = 1, 5, 9,
	    4, 7, 8,
	    2, 3, 6;
//	cin >> A;

	bool b = EVD(A, re, im, L, R);

	cout << "b:" << b << endl;
	cout << "re:" << endl << re;
	cout << "im:" << endl << im;
	cout << "R:" << endl << R;
	cout << "L:" << endl << L;
	
	A = 1, 5, 9,
	    5, 7, 8,
	    9, 8, 6;
//	cin >> A;

	b = symmetric_EVD(A, re, R);

	cout << "b:" << b << endl;
	cout << "eigenValues:" << endl << re;
	cout << "eigenVectors:" << endl << R;
}
*/