/*
    This file is part of the Cheap Matrix library
    Copyright (C) 2000  Nicolas Brodu
    nicolas.brodu@free.fr

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public License
    along with this library; see the file COPYING.LIB.  If not, write to
    the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
    Boston, MA 02111-1307, USA.
*/

#include "CM_annealing.h"
#include "CM_util.h"

namespace CheapMatrix {

Vector Annealing::default_options()
{
	Vector def(5);
	def = 100, -1, 0.999, -1, 1;
	return def;
}

MatrixType Annealing::algorithm(OptimizeTarget& opta, const Matrix& data, const Matrix& target, const Vector& options)
{
	// Get options. We are sure we have at least the default
	int niter = (int)options(1);
	MatrixType temperature = options(2);
	MatrixType ratio = options(3);
	MatrixType step_std = options(4);
	MatrixType constant = options(5);
	// Get the reference to the values. Gradients aren't used
	Vector& values = values_ref(opta);
	// Save current values (make a copy)
	Vector vsave = opta.copy_values();
	const int total_dim = values.rows();
	// Use default if necessary
	if (step_std==-1) step_std = 1.0 / total_dim;

	MatrixType olderror = opta.evaluate(data, target, OptimizeTarget::EVALUATE_ERROR);
	MatrixType newerror = olderror;
	MatrixType min_error = olderror;
	// min error values
	Vector min_values = opta.copy_values();

	// Try to find an initial temperature if none given
	if (temperature==-1) {
		// Calculate 'typical' error variation
		MatrixType max_error=olderror;
		for (int i=0; i<10; i++) {
			values += randn(total_dim) * step_std;
			newerror = opta.evaluate(data, target, OptimizeTarget::EVALUATE_ERROR);
			if (logstream)
				*logstream << "Annealing: Initialization "<<i+1<<", error="<<newerror<<endl;
			if (newerror>max_error) max_error=newerror;
			if (newerror<min_error) min_error=newerror;
			values.copy(vsave);
		}
		// max error step will have 20% chance to be accepted to begin with
		temperature = (max_error - min_error) / (5.0*constant*step_std);
	}
	if (logstream)
		*logstream << "Annealing: Initial temperature="<<temperature<<endl;

	MatrixType step_correction = temperature / step_std;

	// Main loop
	for (int i=1; i<=niter; i++) {
		// try a random step
		values += randn(total_dim) * temperature / step_correction;
		newerror = opta.evaluate(data, target, OptimizeTarget::EVALUATE_ERROR);
		if (logstream)
			*logstream << "Annealing: Step "<<i<<", error="<<newerror<<flush;
		if (randu()< ::exp(-(newerror-olderror)/(constant*temperature*step_std))) {
			vsave = opta.copy_values();
			temperature *= ratio;
			if (logstream) *logstream << ", accepted. Temperature="<<temperature<<flush;
			// Auto-save
			if (!autosavefile.empty() && newerror<min_error) {
				opta.save(autosavefile);
				min_error = newerror;
				min_values.copy(values);
			}
			olderror = newerror;
		}
		else values.copy(vsave);
		if (logstream) *logstream << endl;
	}

	// set result to the minimum found
	values.copy(min_values);
	return newerror;
}

}
