/*
    This file is part of the Cheap Matrix library
    Copyright (C) 2000  Nicolas Brodu
    nicolas.brodu@free.fr

    The function for gradient checking is inspired from a Netlab,
    a Matlab neural network toolbox written by
    Ian T Nabney & Christopher M Bishop (1996, 1997, 1998)
    It is available at http://www.ncrg.aston.ac.uk/netlab/index.html

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public License
    along with this library; see the file COPYING.LIB.  If not, write to
    the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
    Boston, MA 02111-1307, USA.
*/

#include "CM_util.h"
#include "CM_optimizer.h"

namespace CheapMatrix {

OptimizeTarget::~OptimizeTarget()
{
}

// Can save the values. Returns success state
bool OptimizeTarget::save(const string& filename)
{
	if (filename.empty()) return false;
	return ::CheapMatrix::save(values, filename);
}

// Can load the values. Returns success state
bool OptimizeTarget::load(const string& filename)
{
	if (filename.empty()) return false;
	Matrix loadedValues;
	if (!::CheapMatrix::load(loadedValues, filename)) return false;
	if (loadedValues.cols()!=values.cols()
		|| loadedValues.rows()!=values.rows()) return false;
	values.copy(loadedValues); // keep reference right for the aliases
	return true;
}

// Check the evaluate algorithm implementation using finite difference
// Returns the gradients by 'evaluate' in the first column, the finite
// difference estimation in the second, their difference in the third,
// and the absolute value of the relative error commited in the fourth.
// Function inspired from Netlab, a Matlab neural network toolbox written by
// Ian T Nabney & Christopher M Bishop (1996, 1997, 1998)
Matrix OptimizeTarget::gradcheck(const Matrix& data, const Matrix& target)
{
	int total_dim = values.rows();
	if (!total_dim) return Matrix(); // error
	MatrixType epsilon = 1.0e-5;
	Matrix ret = zeros(total_dim,4);
	// Save current values (make a copy)
	Vector xsave = copy_values();
	// Get the gradients
	evaluate(data, target, EVALUATE_GRADIENT);
	ret(1,total_dim,1,1).copy(gradients);
	// Finite difference
	for (int i=1; i<=total_dim; i++) {
		// Move a small way in the ith coordinate
		values(i) += epsilon;
		MatrixType fplus = evaluate(data, target, EVALUATE_ERROR);
		values(i) = xsave(i) - epsilon;
		MatrixType fminus = evaluate(data, target, EVALUATE_ERROR);
		// Use central difference formula for approximation
		ret(i,2) = 0.5*(fplus - fminus)/epsilon;
		values(i) = xsave(i); // restore value from copy
	}
	// diff
	ret(1,total_dim,3,3).copy(ret(1,total_dim,1,1) - ret(1,total_dim,2,2));
	// abs error
	ret(1,total_dim,4,4).copy(fabs(divide(ret(1,total_dim,3,3), ret(1,total_dim,1,1))));
	return ret;
}


Optimizer::~Optimizer() {}

// Convenience functions to propagate friendship to our children
// Reminder: No shallow copy for the values and gradients.
// => values.copy(something) is OK, values = something_else is not.
// The reason is that there might be references in opta.
// This is a choice, of either this, or a lot of copies in the optimizer
Vector& Optimizer::values_ref(OptimizeTarget& opta)
{
	return opta.values;
}
Vector& Optimizer::gradients_ref(OptimizeTarget& opta)
{
	return opta.gradients;
}

MatrixType Optimizer::optimize(OptimizeTarget& opta, const Matrix& data, const Matrix& target, const Vector& options)
{
	Vector def = default_options();
	int def_dim = def.rows();
	// check we have the correct number of options
	if (options.rows() < def_dim) {
		Vector my_copy = def.copy(); // have our own copy, just in case...
		my_copy(1,options.rows()).copy(options); // override defaults
		return algorithm(opta, data, target, my_copy);
	}
	return algorithm(opta, data, target, options);
}

MatrixType Optimizer::optimize(OptimizeTarget& opta, const Matrix& data, const Matrix& target)
{
	Vector def = default_options();
	int def_dim = def.rows();
	// check we have the correct number of options
	if (alg_options.rows() < def_dim) {
		Vector my_copy = def.copy(); // have our own copy, just in case...
		my_copy(1,alg_options.rows()).copy(alg_options); // override defaults
		return algorithm(opta, data, target, my_copy);
	}
	return algorithm(opta, data, target, alg_options);
}

// Let's not define any symbol for this (a static would also work)
// so that we can change it without breaking binary compatibility...
#define UPPER_OPTION_LIMIT 10000

// Facility to set some option and keep the defaults
MatrixType& Optimizer::option(int i)
{
	// get rid of this bothering case
	assert(i>=1 && i<=UPPER_OPTION_LIMIT); // You can increase that if needed
	int cur_dim = alg_options.rows();
	if (i<=cur_dim) return alg_options(i); // we have it, fine
	// So we have to provide a new placeholder
	Vector def = default_options();
	int def_dim = def.rows();
	if (!def_dim) { // agh, no default.
		Vector tmp = Vector(i+10); // let's plan for a future increase...
		if (cur_dim) tmp(1,cur_dim).copy(alg_options); // copy what we have
		alg_options = tmp;     // update
		return alg_options(i); // and return
	}
	// no need to allocate unused fields, and that way it has no effect
	if (i>def_dim) i = def_dim + 1;
	if (cur_dim>def_dim) cur_dim = def_dim; // user set too much!
	Vector tmp = Vector(def_dim+1);
	if (cur_dim) tmp(1,cur_dim).copy(alg_options); // copy what we have
	// and what we didn't have...
	if (cur_dim<def_dim) tmp(cur_dim+1,def_dim).copy(def(cur_dim+1,def_dim));
	alg_options = tmp;     // update
	return alg_options(i); // and return
}

Vector Optimizer::default_options()
{
	return Vector();
}

}
